/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2018 Marco Martin <mart@kde.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#ifndef KWIN_VIRTUALDESKTOPS_DBUS_TYPES_H
#define KWIN_VIRTUALDESKTOPS_DBUS_TYPES_H

#include <QtDBus>

namespace KWin
{

struct DBusDesktopDataStruct {
    uint position;
    QString id;
    QString name;
};
typedef QVector<DBusDesktopDataStruct> DBusDesktopDataVector;
}

const QDBusArgument &operator<<(QDBusArgument &argument, const KWin::DBusDesktopDataStruct &desk);
const QDBusArgument &operator>>(const QDBusArgument &argument, KWin::DBusDesktopDataStruct &desk);

Q_DECLARE_METATYPE(KWin::DBusDesktopDataStruct)

const QDBusArgument &operator<<(QDBusArgument &argument, const KWin::DBusDesktopDataVector &deskVector);
const QDBusArgument &operator>>(const QDBusArgument &argument, KWin::DBusDesktopDataVector &deskVector);

Q_DECLARE_METATYPE(KWin::DBusDesktopDataVector)

#endif // KWIN_VIRTUALDESKTOPS_DBUS_TYPES_H
