/********************************************************************
KWin - the KDE window manager
This file is part of the KDE project.

Copyright (C) 2009 Martin Gräßlin <mgraesslin@kde.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "cube_proxy.h"
#include "cube.h"
#include "cube_inside.h"

namespace KWin
{

CubeEffectProxy::CubeEffectProxy(CubeEffect* effect)
    : m_effect(effect)
{
}

CubeEffectProxy::~CubeEffectProxy()
{
}

void CubeEffectProxy::registerCubeInsideEffect(CubeInsideEffect* effect)
{
    m_effect->registerCubeInsideEffect(effect);
}

void CubeEffectProxy::unregisterCubeInsideEffect(CubeInsideEffect* effect)
{
    m_effect->unregisterCubeInsideEffect(effect);
}

} // namespace
