/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2018 Martin Flöser <mgraesslin@kde.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/
#pragma once
#include "input_event_spy.h"

namespace KWin
{

class TouchHideCursorSpy : public InputEventSpy
{
public:
    void pointerEvent(KWin::MouseEvent *event) override;
    void wheelEvent(KWin::WheelEvent *event) override;
    void touchDown(qint32 id, const QPointF &pos, quint32 time) override;

private:
    void showCursor();
    void hideCursor();

    bool m_cursorHidden = false;
};

}
