/*
 * Copyright (c) 2004 Lubos Lunak <l.lunak@kde.org>
 * Copyright (c) 2020 Ismael Asensio <isma.af@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef KWIN_RULESDIALOG_H
#define KWIN_RULESDIALOG_H

#include "rulesmodel.h"
#include "../../rules.h"

#include <QDialog>

namespace KWin
{

class Rules;

class RulesDialog : public QDialog
{
    Q_OBJECT

public:
    explicit RulesDialog(QWidget* parent = nullptr, const char* name = nullptr);

    Rules* edit(Rules* r, const QVariantMap& info, bool show_hints);

protected:
    void accept() override;

private:
    RulesModel* m_rulesModel;
    QWidget *m_quickWidget;
    Rules* m_rules;
};

} // namespace

#endif // KWIN_RULESDIALOG_H
