/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2016 Roman Gilg <subdiff@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/
#ifndef KWIN_DRM_OBJECT_CONNECTOR_H
#define KWIN_DRM_OBJECT_CONNECTOR_H

#include "drm_object.h"

namespace KWin
{

class DrmConnector : public DrmObject
{
public:
    DrmConnector(uint32_t connector_id, int fd);

    ~DrmConnector() override;

    bool atomicInit() override;

    enum class PropertyIndex {
        CrtcId = 0,
        Count
    };

    QVector<uint32_t> encoders() {
        return m_encoders;
    }
    
    bool initProps() override;
    bool isConnected();


private:
    QVector<uint32_t> m_encoders;
};

}

#endif

